@@ -92,6 +92,11 @@ typedef int64_t             s8;
  */
 #define DEX_MAGIC_VERS_37  "037\0"
 
+/* The version for android O, encoded in 4 bytes of ASCII. This differentiates dex files that may
+ * contain invoke-custom, invoke-polymorphic, call-sites, and method handles.
+ */
+#define DEX_MAGIC_VERS_38  "038\0"
+
 /* current version, encoded in 4 bytes of ASCII */
 #define DEX_MAGIC_VERS  "036\0"
 
@@ -191,6 +196,8 @@ enum {
     kDexAnnotationLong          = 0x06,
     kDexAnnotationFloat         = 0x10,
     kDexAnnotationDouble        = 0x11,
+    kDexAnnotationMethodType    = 0x15,
+    kDexAnnotationMethodHandle  = 0x16,
     kDexAnnotationString        = 0x17,
     kDexAnnotationType          = 0x18,
     kDexAnnotationField         = 0x19,
@@ -214,6 +221,8 @@ enum {
     kDexTypeFieldIdItem              = 0x0004,
     kDexTypeMethodIdItem             = 0x0005,
     kDexTypeClassDefItem             = 0x0006,
+    kDexTypeCallSiteIdItem           = 0x0007,
+    kDexTypeMethodHandleItem         = 0x0008,
     kDexTypeMapList                  = 0x1000,
     kDexTypeTypeList                 = 0x1001,
     kDexTypeAnnotationSetRefList     = 0x1002,
@@ -354,6 +363,38 @@ struct DexClassDef {
     u4  staticValuesOff;    /* file offset to DexEncodedArray */
 };
 
+/*
+ * Direct-mapped "call_site_id_item"
+ */
+struct DexCallSiteId {
+    u4  callSiteOff;        /* file offset to DexEncodedArray */
+};
+
+/*
+ * Enumeration of method handle type codes.
+ */
+enum MethodHandleType {
+    STATIC_PUT = 0x00,
+    STATIC_GET = 0x01,
+    INSTANCE_PUT = 0x02,
+    INSTANCE_GET = 0x03,
+    INVOKE_STATIC = 0x04,
+    INVOKE_INSTANCE = 0x05,
+    INVOKE_CONSTRUCTOR = 0x06,
+    INVOKE_DIRECT = 0x07,
+    INVOKE_INTERFACE = 0x08
+};
+
+/*
+ * Direct-mapped "method_handle_item"
+ */
+struct DexMethodHandleItem {
+    u2 methodHandleType;    /* type of method handle */
+    u2 reserved1;           /* reserved for future use */
+    u2 fieldOrMethodIdx;    /* index of associated field or method */
+    u2 reserved2;           /* reserved for future use */
+};
+
 /*
  * Direct-mapped "type_item".
  */
@@ -606,7 +647,7 @@ int dexSwapAndVerify(u1* addr, int len);
  *
  * Return 0 on success.
  */
-int dexSwapAndVerifyIfNecessary(u1* addr, int len);
+int dexSwapAndVerifyIfNecessary(u1* addr, size_t len);
 
 /*
  * Check to see if the file magic and format version in the given
