@@ -45,10 +45,12 @@
 struct CheckState {
     const DexHeader*  pHeader;
     const u1*         fileStart;
-    const u1*         fileEnd;      // points to fileStart + fileLen
+    const u1*         fileEnd;             // points to fileStart + fileLen
     u4                fileLen;
-    DexDataMap*       pDataMap;     // set after map verification
-    const DexFile*    pDexFile;     // set after intraitem verification
+    DexDataMap*       pDataMap;            // set after map verification
+    const DexFile*    pDexFile;            // set after intraitem verification
+    const DexMapItem* pCallSiteIds;        // set after intraitem verification
+    const DexMapItem* pMethodHandleItems;  // set after intraitem verification
 
     /*
      * bitmap of type_id indices that have been used to define classes;
@@ -335,6 +337,8 @@ static u4 mapTypeToBitMask(int mapType) {
         case kDexTypeAnnotationItem:           return 1 << 15;
         case kDexTypeEncodedArrayItem:         return 1 << 16;
         case kDexTypeAnnotationsDirectoryItem: return 1 << 17;
+        case kDexTypeCallSiteIdItem:           return 1 << 18;
+        case kDexTypeMethodHandleItem:         return 1 << 19;
         default: {
             ALOGE("Unknown map item type %04x", mapType);
             return 0;
@@ -430,6 +434,12 @@ static bool swapMap(CheckState* state, DexMapList* pMap)
             return false;
         }
 
+        if (item->type == kDexTypeCallSiteIdItem) {
+            state->pCallSiteIds = item;
+        } else if (item->type == kDexTypeMethodHandleItem) {
+            state->pMethodHandleItems = item;
+        }
+
         usedBits |= bit;
         lastOffset = item->offset;
         item++;
@@ -1036,6 +1046,56 @@ static void* crossVerifyClassDefItem(const CheckState* state, void* ptr) {
     return (void*) (item + 1);
 }
 
+/* Perform cross-item verification of call_site_id. */
+static void* crossVerifyCallSiteId(const CheckState* state, void* ptr) {
+    const DexCallSiteId* item = (const DexCallSiteId*) ptr;
+    if (state->pCallSiteIds == nullptr) {
+        ALOGE("Verifying call site but expecting none");
+        return NULL;
+    }
+    if (item->callSiteOff < state->pHeader->dataOff ||
+        item->callSiteOff >= state->pHeader->dataOff + state->pHeader->dataSize) {
+        ALOGE("Bad call site offset: %u", item->callSiteOff);
+        return NULL;
+    }
+    return (void*) (item + 1);
+}
+
+/* Perform cross-item verification of method_handle_item. */
+static void* crossVerifyMethodHandleItem(const CheckState* state, void* ptr) {
+    const DexMethodHandleItem* item = (const DexMethodHandleItem*) ptr;
+    if (state->pMethodHandleItems == nullptr) {
+        ALOGE("Verifying method handle but expecting none");
+        return NULL;
+    }
+    if (item->methodHandleType > (u2) MethodHandleType::INVOKE_INTERFACE) {
+        ALOGE("Unknown method handle type: %u", item->methodHandleType);
+        return NULL;
+    }
+    switch ((MethodHandleType) item->methodHandleType) {
+        case MethodHandleType::STATIC_PUT:
+        case MethodHandleType::STATIC_GET:
+        case MethodHandleType::INSTANCE_PUT:
+        case MethodHandleType::INSTANCE_GET:
+            if (item->fieldOrMethodIdx >= state->pHeader->fieldIdsSize) {
+                ALOGE("Method handle has invalid field id: %u\n", item->fieldOrMethodIdx);
+                return NULL;
+            }
+            break;
+        case MethodHandleType::INVOKE_STATIC:
+        case MethodHandleType::INVOKE_INSTANCE:
+        case MethodHandleType::INVOKE_CONSTRUCTOR:
+        case MethodHandleType::INVOKE_DIRECT:
+        case MethodHandleType::INVOKE_INTERFACE:
+            if (item->fieldOrMethodIdx >= state->pHeader->methodIdsSize) {
+                ALOGE("Method handle has invalid method id: %u\n", item->fieldOrMethodIdx);
+                return NULL;
+            }
+            break;
+    }
+    return (void*) (item + 1);
+}
+
 /* Helper for swapAnnotationsDirectoryItem(), which performs
  * byte-swapping and intra-item verification on an
  * annotation_directory_item's field elements. */
@@ -1164,6 +1224,26 @@ static void* swapAnnotationsDirectoryItem(const CheckState* state, void* ptr) {
     return addr;
 }
 
+static void* swapCallSiteId(const CheckState* state, void* ptr) {
+    DexCallSiteId* item = (DexCallSiteId*) ptr;
+
+    CHECK_PTR_RANGE(item, item + 1);
+    SWAP_OFFSET4(item->callSiteOff);
+
+    return (item + 1);
+}
+
+static void* swapMethodHandleItem(const CheckState* state, void* ptr) {
+    DexMethodHandleItem* item = (DexMethodHandleItem*) ptr;
+
+    CHECK_PTR_RANGE(item, item + 1);
+    SWAP_FIELD2(item->methodHandleType);
+    SWAP_FIELD2(item->fieldOrMethodIdx);
+
+    return (item + 1);
+}
+
+
 /* Helper for crossVerifyAnnotationsDirectoryItem(), which checks the
  * field elements. */
 static const u1* crossVerifyFieldAnnotations(const CheckState* state, u4 count,
@@ -2144,6 +2224,13 @@ static const u1* verifyEncodedArray(const CheckState* state,
     return data;
 }
 
+static u4 numberOfMethodHandles(const CheckState* state) {
+    if (state->pMethodHandleItems != nullptr) {
+        return state->pMethodHandleItems->size;
+    }
+    return 0;
+}
+
 /* Helper for *VerifyAnnotationItem() and *VerifyEncodedArrayItem(), which
  * verifies an encoded_value. */
 static const u1* verifyEncodedValue(const CheckState* state,
@@ -2186,6 +2273,24 @@ static const u1* verifyEncodedValue(const CheckState* state,
             data += valueArg + 1;
             break;
         }
+        case kDexAnnotationMethodType: {
+            if (valueArg > 3) {
+                ALOGE("Bogus method type size %#x", valueArg);
+                return NULL;
+            }
+            u4 idx = readUnsignedLittleEndian(state, &data, valueArg + 1);
+            CHECK_INDEX(idx, state->pHeader->protoIdsSize);
+            break;
+        }
+        case kDexAnnotationMethodHandle: {
+            if (valueArg > 3) {
+                ALOGE("Bogus method type size %#x", valueArg);
+                return NULL;
+            }
+            u4 idx = readUnsignedLittleEndian(state, &data, valueArg + 1);
+            CHECK_INDEX(idx, numberOfMethodHandles(state));
+            break;
+        }
         case kDexAnnotationString: {
             if (valueArg > 3) {
                 ALOGE("Bogus string size %#x", valueArg);
@@ -2595,6 +2700,18 @@ static bool swapEverythingButHeaderAndMap(CheckState* state,
                         sizeof(u4), &lastOffset);
                 break;
             }
+            case kDexTypeCallSiteIdItem: {
+                okay = checkBoundsAndIterateSection(state, sectionOffset,
+                        sectionCount, sectionOffset, sectionCount,
+                        swapCallSiteId, sizeof(u4), &lastOffset);
+                break;
+            }
+            case kDexTypeMethodHandleItem: {
+                okay = checkBoundsAndIterateSection(state, sectionOffset,
+                        sectionCount, sectionOffset, sectionCount,
+                        swapMethodHandleItem, sizeof(u4), &lastOffset);
+                break;
+            }
             case kDexTypeMapList: {
                 /*
                  * The map section was swapped early on, but do some
@@ -2742,6 +2859,16 @@ static bool crossVerifyEverything(CheckState* state, DexMapList* pMap)
                 state->pDefinedClassBits = NULL;
                 break;
             }
+            case kDexTypeCallSiteIdItem: {
+                okay = iterateSection(state, sectionOffset, sectionCount,
+                        crossVerifyCallSiteId, sizeof(u4), NULL);
+                break;
+            }
+            case kDexTypeMethodHandleItem: {
+                okay = iterateSection(state, sectionOffset, sectionCount,
+                        crossVerifyMethodHandleItem, sizeof(u4), NULL);
+                break;
+            }
             case kDexTypeAnnotationSetRefList: {
                 okay = iterateSection(state, sectionOffset, sectionCount,
                         crossVerifyAnnotationSetRefList, sizeof(u4), NULL);
@@ -2792,8 +2919,9 @@ bool dexHasValidMagic(const DexHeader* pHeader)
     }
 
     if ((memcmp(version, DEX_MAGIC_VERS, 4) != 0) &&
-            (memcmp(version, DEX_MAGIC_VERS_API_13, 4) != 0) &&
-            (memcmp(version, DEX_MAGIC_VERS_37, 4) != 0)) {
+        (memcmp(version, DEX_MAGIC_VERS_API_13, 4) != 0) &&
+        (memcmp(version, DEX_MAGIC_VERS_37, 4) != 0) &&
+        (memcmp(version, DEX_MAGIC_VERS_38, 4) != 0)) {
         /*
          * Magic was correct, but this is an unsupported older or
          * newer format variant.
@@ -2813,7 +2941,7 @@ bool dexHasValidMagic(const DexHeader* pHeader)
  *
  * Returns 0 on success, nonzero on failure.
  */
-int dexSwapAndVerify(u1* addr, int len)
+int dexSwapAndVerify(u1* addr, size_t len)
 {
     DexHeader* pHeader;
     CheckState state;
@@ -2833,14 +2961,10 @@ int dexSwapAndVerify(u1* addr, int len)
     }
 
     if (okay) {
-        int expectedLen = (int) SWAP4(pHeader->fileSize);
-        if (len < expectedLen) {
-            ALOGE("ERROR: Bad length: expected %d, got %d", expectedLen, len);
+        u4 expectedLen = SWAP4(pHeader->fileSize);
+        if (len != expectedLen) {
+            ALOGE("ERROR: Bad length: expected %u, got %zu", expectedLen, len);
             okay = false;
-        } else if (len != expectedLen) {
-            ALOGW("WARNING: Odd length: expected %d, got %d", expectedLen,
-                    len);
-            // keep going
         }
     }
 
@@ -2939,7 +3063,7 @@ int dexSwapAndVerify(u1* addr, int len)
  *
  * Returns 0 on success, nonzero on failure.
  */
-int dexSwapAndVerifyIfNecessary(u1* addr, int len)
+int dexSwapAndVerifyIfNecessary(u1* addr, size_t len)
 {
     if (memcmp(addr, DEX_OPT_MAGIC, 4) == 0) {
         // It is an optimized dex file.
