@@ -16,11 +16,15 @@
 
 package com.android.dx.cf.code;
 
+import com.android.dex.DexFormat;
+import com.android.dx.dex.DexOptions;
 import com.android.dx.rop.code.LocalItem;
 import com.android.dx.rop.cst.Constant;
+import com.android.dx.rop.cst.CstCallSiteRef;
 import com.android.dx.rop.cst.CstFieldRef;
 import com.android.dx.rop.cst.CstInteger;
 import com.android.dx.rop.cst.CstInterfaceMethodRef;
+import com.android.dx.rop.cst.CstInvokeDynamic;
 import com.android.dx.rop.cst.CstMethodRef;
 import com.android.dx.rop.cst.CstType;
 import com.android.dx.rop.type.Prototype;
@@ -56,13 +60,17 @@ public class Simulator {
     /** {@code non-null;} visitor instance to use */
     private final SimVisitor visitor;
 
+    /** {@code non-null;} options for dex output */
+    private final DexOptions dexOptions;
+
     /**
      * Constructs an instance.
      *
      * @param machine {@code non-null;} machine to use when simulating
      * @param method {@code non-null;} method data to use
+     * @param dexOptions {@code non-null;} options for dex output
      */
-    public Simulator(Machine machine, ConcreteMethod method) {
+    public Simulator(Machine machine, ConcreteMethod method, DexOptions dexOptions) {
         if (machine == null) {
             throw new NullPointerException("machine == null");
         }
@@ -75,6 +83,7 @@ public class Simulator {
         this.code = method.getCode();
         this.localVariables = method.getLocalVariables();
         this.visitor = new SimVisitor();
+        this.dexOptions = dexOptions;
     }
 
     /**
@@ -105,7 +114,7 @@ public class Simulator {
      * Simulates the effect of the instruction at the given offset, by
      * making appropriate calls on the given frame.
      *
-     * @param offset {@code >= 0;} offset of the instruction to simulate
+     * @param offset {@code offset >= 0;} offset of the instruction to simulate
      * @param frame {@code non-null;} frame to operate on
      * @return the length of the instruction, in bytes
      */
@@ -581,8 +590,9 @@ public class Simulator {
                 localType = local.getType();
                 if (localType.getBasicFrameType() !=
                         type.getBasicFrameType()) {
-                    BaseMachine.throwLocalMismatch(type, localType);
-                    return;
+                    // wrong type, ignore local variable info
+                    local = null;
+                    localType = type;
                 }
             } else {
                 localType = type;
@@ -647,33 +657,67 @@ public class Simulator {
                     machine.popArgs(frame, Type.OBJECT, fieldType);
                     break;
                 }
-                case ByteOps.INVOKEINTERFACE: {
+                case ByteOps.INVOKEINTERFACE:
+                case ByteOps.INVOKEVIRTUAL:
+                case ByteOps.INVOKESPECIAL:
+                case ByteOps.INVOKESTATIC: {
                     /*
                      * Convert the interface method ref into a normal
-                     * method ref.
+                     * method ref if necessary.
                      */
-                    cst = ((CstInterfaceMethodRef) cst).toMethodRef();
-                    // and fall through...
-                }
-                case ByteOps.INVOKEVIRTUAL:
-                case ByteOps.INVOKESPECIAL: {
+                    if (cst instanceof CstInterfaceMethodRef) {
+                        if (opcode != ByteOps.INVOKEINTERFACE) {
+                            if (!dexOptions.canUseDefaultInterfaceMethods()) {
+                                throw new SimException(
+                                    "default or static interface method used without " +
+                                    "--min-sdk-version >= " + DexFormat.API_DEFAULT_INTERFACE_METHODS);
+                            }
+                        }
+                        cst = ((CstInterfaceMethodRef) cst).toMethodRef();
+                    }
+
                     /*
-                     * Get the instance prototype, and use it to direct
-                     * the machine.
+                     * Check whether invoke-polymorphic is required and supported.
+                    */
+                    if (cst instanceof CstMethodRef) {
+                        CstMethodRef methodRef = (CstMethodRef) cst;
+                        if (methodRef.isSignaturePolymorphic()) {
+                            if (!dexOptions.canUseInvokePolymorphic()) {
+                                throw new SimException(
+                                    "signature-polymorphic method called without " +
+                                    "--min-sdk-version >= " + DexFormat.API_INVOKE_POLYMORPHIC);
+                            }
+                            if (opcode != ByteOps.INVOKEVIRTUAL) {
+                                throw new SimException(
+                                    "Unsupported signature polymorphic invocation (" +
+                                    ByteOps.opName(opcode) + ")");
+                            }
+                        }
+                    }
+
+                    /*
+                     * Get the instance or static prototype, and use it to
+                     * direct the machine.
                      */
+                    boolean staticMethod = (opcode == ByteOps.INVOKESTATIC);
                     Prototype prototype =
-                        ((CstMethodRef) cst).getPrototype(false);
+                        ((CstMethodRef) cst).getPrototype(staticMethod);
                     machine.popArgs(frame, prototype);
                     break;
                 }
-                case ByteOps.INVOKESTATIC: {
-                    /*
-                     * Get the static prototype, and use it to direct
-                     * the machine.
-                     */
-                    Prototype prototype =
-                        ((CstMethodRef) cst).getPrototype(true);
+                case ByteOps.INVOKEDYNAMIC: {
+                    if (!dexOptions.canUseInvokeCustom()) {
+                        throw new SimException(
+                            "invalid opcode " + Hex.u1(opcode) +
+                            " (invokedynamic requires --min-sdk-version >= " +
+                            DexFormat.API_INVOKE_POLYMORPHIC + ")");
+                    }
+                    CstInvokeDynamic invokeDynamicRef = (CstInvokeDynamic) cst;
+                    Prototype prototype = invokeDynamicRef.getPrototype();
                     machine.popArgs(frame, prototype);
+                    // Change the constant to be associated with instruction to
+                    // a call site reference.
+                    cst = invokeDynamicRef.addReference();
                     break;
                 }
                 case ByteOps.MULTIANEWARRAY: {
