@@ -27,12 +27,10 @@ import com.android.dx.cf.direct.ClassPathOpener.FileNameFilter;
 import com.android.dx.cf.direct.DirectClassFile;
 import com.android.dx.cf.direct.StdAttributeFactory;
 import com.android.dx.cf.iface.ParseException;
-import com.android.dx.command.DxConsole;
 import com.android.dx.command.UsageException;
 import com.android.dx.dex.DexOptions;
 import com.android.dx.dex.cf.CfOptions;
 import com.android.dx.dex.cf.CfTranslator;
-import com.android.dx.dex.cf.CodeStatistics;
 import com.android.dx.dex.code.PositionList;
 import com.android.dx.dex.file.ClassDefItem;
 import com.android.dx.dex.file.DexFile;
@@ -42,8 +40,12 @@ import com.android.dx.merge.DexMerger;
 import com.android.dx.rop.annotation.Annotation;
 import com.android.dx.rop.annotation.Annotations;
 import com.android.dx.rop.annotation.AnnotationsList;
+import com.android.dx.rop.code.RegisterSpec;
 import com.android.dx.rop.cst.CstNat;
 import com.android.dx.rop.cst.CstString;
+import com.android.dx.rop.cst.CstType;
+import com.android.dx.rop.type.Prototype;
+import com.android.dx.rop.type.Type;
 
 import java.io.BufferedReader;
 import java.io.ByteArrayInputStream;
@@ -70,7 +72,6 @@ import java.util.concurrent.ExecutionException;
 import java.util.concurrent.ExecutorService;
 import java.util.concurrent.Executors;
 import java.util.concurrent.Future;
-import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.ThreadPoolExecutor;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicInteger;
@@ -84,6 +85,8 @@ import java.util.jar.Manifest;
  */
 public class Main {
 
+    public static final int CONCURRENCY_LEVEL = 4;
+
     /**
      * File extension of a {@code .dex} file.
      */
@@ -164,74 +167,72 @@ public class Main {
     private static final int MAX_FIELD_ADDED_DURING_DEX_CREATION = 9;
 
     /** number of errors during processing */
-    private static AtomicInteger errors = new AtomicInteger(0);
+    private AtomicInteger errors = new AtomicInteger(0);
 
     /** {@code non-null;} parsed command-line arguments */
-    private static Arguments args;
+    private Arguments args;
 
     /** {@code non-null;} output file in-progress */
-    private static DexFile outputDex;
+    private DexFile outputDex;
 
     /**
      * {@code null-ok;} map of resources to include in the output, or
      * {@code null} if resources are being ignored
      */
-    private static TreeMap<String, byte[]> outputResources;
+    private TreeMap<String, byte[]> outputResources;
 
     /** Library .dex files to merge into the output .dex. */
-    private static final List<byte[]> libraryDexBuffers = new ArrayList<byte[]>();
+    private final List<byte[]> libraryDexBuffers = new ArrayList<byte[]>();
 
     /** Thread pool object used for multi-thread class translation. */
-    private static ExecutorService classTranslatorPool;
+    private ExecutorService classTranslatorPool;
 
     /** Single thread executor, for collecting results of parallel translation,
      * and adding classes to dex file in original input file order. */
-    private static ExecutorService classDefItemConsumer;
+    private ExecutorService classDefItemConsumer;
 
     /** Futures for {@code classDefItemConsumer} tasks. */
-    private static List<Future<Boolean>> addToDexFutures =
+    private List<Future<Boolean>> addToDexFutures =
             new ArrayList<Future<Boolean>>();
 
     /** Thread pool object used for multi-thread dex conversion (to byte array).
      * Used in combination with multi-dex support, to allow outputing
      * a completed dex file, in parallel with continuing processing. */
-    private static ExecutorService dexOutPool;
+    private ExecutorService dexOutPool;
 
     /** Futures for {@code dexOutPool} task. */
-    private static List<Future<byte[]>> dexOutputFutures =
-            new ArrayList<Future<byte[]>>();
+    private List<Future<byte[]>> dexOutputFutures = new ArrayList<Future<byte[]>>();
 
     /** Lock object used to to coordinate dex file rotation, and
      * multi-threaded translation. */
-    private static Object dexRotationLock = new Object();
+    private Object dexRotationLock = new Object();
 
     /** Record the number if method indices "reserved" for files
      * committed to translation in the context of the current dex
      * file, but not yet added. */
-    private static int maxMethodIdsInProcess = 0;
+    private int maxMethodIdsInProcess = 0;
 
     /** Record the number if field indices "reserved" for files
      * committed to translation in the context of the current dex
      * file, but not yet added. */
-    private static int maxFieldIdsInProcess = 0;
+    private int maxFieldIdsInProcess = 0;
 
     /** true if any files are successfully processed */
-    private static volatile boolean anyFilesProcessed;
+    private volatile boolean anyFilesProcessed;
 
     /** class files older than this must be defined in the target dex file. */
-    private static long minimumFileAge = 0;
+    private long minimumFileAge = 0;
 
-    private static Set<String> classesInMainDex = null;
+    private Set<String> classesInMainDex = null;
 
-    private static List<byte[]> dexOutputArrays = new ArrayList<byte[]>();
+    private List<byte[]> dexOutputArrays = new ArrayList<byte[]>();
 
-    private static OutputStreamWriter humanOutWriter = null;
+    private OutputStreamWriter humanOutWriter = null;
 
-    /**
-     * This class is uninstantiable.
-     */
-    private Main() {
-        // This space intentionally left blank.
+    private final DxContext context;
+
+    public Main(DxContext context) {
+        this.context = context;
     }
 
     /**
@@ -239,21 +240,34 @@ public class Main {
      * @param argArray the command line arguments
      */
     public static void main(String[] argArray) throws IOException {
-        Arguments arguments = new Arguments();
+        DxContext context = new DxContext();
+        Arguments arguments = new Arguments(context);
         arguments.parse(argArray);
 
-        int result = run(arguments);
+        int result = new Main(context).runDx(arguments);
+
         if (result != 0) {
             System.exit(result);
         }
     }
 
+    public static void clearInternTables() {
+        Prototype.clearInternTable();
+        RegisterSpec.clearInternTable();
+        CstType.clearInternTable();
+        Type.clearInternTable();
+    }
+
     /**
      * Run and return a result code.
      * @param arguments the data + parameters for the conversion
-     * @return 0 if success > 0 otherwise.
+     * @return 0 if success &gt; 0 otherwise.
      */
     public static int run(Arguments arguments) throws IOException {
+        return new Main(new DxContext()).runDx(arguments);
+    }
+
+    public int runDx(Arguments arguments) throws IOException {
 
         // Reset the error count to start fresh.
         errors.set(0);
@@ -281,24 +295,12 @@ public class Main {
         }
     }
 
-    /**
-     * {@code non-null;} Error message for too many method/field/type ids.
-     */
-    public static String getTooManyIdsErrorMessage() {
-        if (args.multiDex) {
-            return "The list of classes given in " + Arguments.MAIN_DEX_LIST_OPTION +
-                   " is too big and does not fit in the main dex.";
-        } else {
-            return "You may try using " + Arguments.MULTI_DEX_OPTION + " option.";
-        }
-    }
-
-    private static int runMonoDex() throws IOException {
+    private int runMonoDex() throws IOException {
 
         File incrementalOutFile = null;
         if (args.incremental) {
             if (args.outName == null) {
-                System.err.println(
+                context.err.println(
                         "error: no incremental output name specified");
                 return -1;
             }
@@ -352,7 +354,7 @@ public class Main {
         return 0;
     }
 
-    private static int runMultiDex() throws IOException {
+    private int runMultiDex() throws IOException {
 
         assert !args.incremental;
 
@@ -458,7 +460,7 @@ public class Main {
      * @return the bytes of the merged dex file, or null if both the update
      *     and the base dex do not exist.
      */
-    private static byte[] mergeIncremental(byte[] update, File base) throws IOException {
+    private byte[] mergeIncremental(byte[] update, File base) throws IOException {
         Dex dexA = null;
         Dex dexB = null;
 
@@ -478,7 +480,7 @@ public class Main {
         } else if (dexB == null) {
             result = dexA;
         } else {
-            result = new DexMerger(new Dex[] {dexA, dexB}, CollisionPolicy.KEEP_FIRST).merge();
+            result = new DexMerger(new Dex[] {dexA, dexB}, CollisionPolicy.KEEP_FIRST, context).merge();
         }
 
         ByteArrayOutputStream bytesOut = new ByteArrayOutputStream();
@@ -490,7 +492,7 @@ public class Main {
      * Merges the dex files in library jars. If multiple dex files define the
      * same type, this fails with an exception.
      */
-    private static byte[] mergeLibraryDexBuffers(byte[] outArray) throws IOException {
+    private byte[] mergeLibraryDexBuffers(byte[] outArray) throws IOException {
         ArrayList<Dex> dexes = new ArrayList<Dex>();
         if (outArray != null) {
             dexes.add(new Dex(outArray));
@@ -501,7 +503,7 @@ public class Main {
         if (dexes.isEmpty()) {
             return null;
         }
-        Dex merged = new DexMerger(dexes.toArray(new Dex[dexes.size()]), CollisionPolicy.FAIL).merge();
+        Dex merged = new DexMerger(dexes.toArray(new Dex[dexes.size()]), CollisionPolicy.FAIL, context).merge();
         return merged.getBytes();
     }
 
@@ -511,7 +513,7 @@ public class Main {
      *
      * @return whether processing was successful
      */
-    private static boolean processAllFiles() {
+    private boolean processAllFiles() {
         createDexFile();
 
         if (args.jarOutput) {
@@ -596,10 +598,10 @@ public class Main {
                     int count = errors.incrementAndGet();
                     if (count < 10) {
                         if (args.debug) {
-                            DxConsole.err.println("Uncaught translation error:");
-                            ex.getCause().printStackTrace(DxConsole.err);
+                            context.err.println("Uncaught translation error:");
+                            ex.getCause().printStackTrace(context.err);
                         } else {
-                            DxConsole.err.println("Uncaught translation error: " + ex.getCause());
+                            context.err.println("Uncaught translation error: " + ex.getCause());
                         }
                     } else {
                         throw new InterruptedException("Too many errors");
@@ -614,13 +616,13 @@ public class Main {
         } catch (Exception e) {
             classTranslatorPool.shutdownNow();
             classDefItemConsumer.shutdownNow();
-            e.printStackTrace(System.out);
+            e.printStackTrace(context.out);
             throw new RuntimeException("Unexpected exception in translator thread.", e);
         }
 
         int errorNum = errors.get();
         if (errorNum != 0) {
-            DxConsole.err.println(errorNum + " error" +
+            context.err.println(errorNum + " error" +
                     ((errorNum == 1) ? "" : "s") + "; aborting");
             return false;
         }
@@ -630,18 +632,18 @@ public class Main {
         }
 
         if (!(anyFilesProcessed || args.emptyOk)) {
-            DxConsole.err.println("no classfiles specified");
+            context.err.println("no classfiles specified");
             return false;
         }
 
         if (args.optimize && args.statistics) {
-            CodeStatistics.dumpStatistics(DxConsole.out);
+            context.codeStatistics.dumpStatistics(context.out);
         }
 
         return true;
     }
 
-    private static void createDexFile() {
+    private void createDexFile() {
         outputDex = new DexFile(args.dexOptions);
 
         if (args.dumpWidth != 0) {
@@ -649,7 +651,7 @@ public class Main {
         }
     }
 
-    private static void rotateDexFile() {
+    private void rotateDexFile() {
         if (outputDex != null) {
             if (dexOutPool != null) {
                 dexOutputFutures.add(dexOutPool.submit(new DexWriter(outputDex)));
@@ -669,7 +671,7 @@ public class Main {
      * containing class files.
      * @param filter {@code non-null;} A filter for excluding files.
      */
-    private static void processOne(String pathname, FileNameFilter filter) {
+    private void processOne(String pathname, FileNameFilter filter) {
         ClassPathOpener opener;
 
         opener = new ClassPathOpener(pathname, true, filter, new FileBytesConsumer());
@@ -679,7 +681,7 @@ public class Main {
         }
     }
 
-    private static void updateStatus(boolean res) {
+    private void updateStatus(boolean res) {
         anyFilesProcessed |= res;
     }
 
@@ -691,7 +693,7 @@ public class Main {
      * @param bytes {@code non-null;} contents of the file
      * @return whether processing was successful
      */
-    private static boolean processFileBytes(String name, long lastModified, byte[] bytes) {
+    private boolean processFileBytes(String name, long lastModified, byte[] bytes) {
 
         boolean isClass = name.endsWith(".class");
         boolean isClassesDex = name.equals(DexFormat.DEX_IN_JAR_NAME);
@@ -699,13 +701,13 @@ public class Main {
 
         if (!isClass && !isClassesDex && !keepResources) {
             if (args.verbose) {
-                DxConsole.out.println("ignored resource " + name);
+                context.out.println("ignored resource " + name);
             }
             return false;
         }
 
         if (args.verbose) {
-            DxConsole.out.println("processing " + name + "...");
+            context.out.println("processing " + name + "...");
         }
 
         String fixedName = fixPath(name);
@@ -745,7 +747,7 @@ public class Main {
      * @param bytes {@code non-null;} contents of the file
      * @return whether processing was successful
      */
-    private static boolean processClass(String name, byte[] bytes) {
+    private boolean processClass(String name, byte[] bytes) {
         if (! args.coreLibrary) {
             checkClassName(name);
         }
@@ -764,7 +766,7 @@ public class Main {
     }
 
 
-    private static DirectClassFile parseClass(String name, byte[] bytes) {
+    private DirectClassFile parseClass(String name, byte[] bytes) {
 
         DirectClassFile cf = new DirectClassFile(bytes, name,
                 args.cfOptions.strictNameCheck);
@@ -773,23 +775,23 @@ public class Main {
         return cf;
     }
 
-    private static ClassDefItem translateClass(byte[] bytes, DirectClassFile cf) {
+    private ClassDefItem translateClass(byte[] bytes, DirectClassFile cf) {
         try {
-            return CfTranslator.translate(cf, bytes, args.cfOptions,
+            return CfTranslator.translate(context, cf, bytes, args.cfOptions,
                     args.dexOptions, outputDex);
         } catch (ParseException ex) {
-            DxConsole.err.println("\ntrouble processing:");
+            context.err.println("\ntrouble processing:");
             if (args.debug) {
-                ex.printStackTrace(DxConsole.err);
+                ex.printStackTrace(context.err);
             } else {
-                ex.printContext(DxConsole.err);
+                ex.printContext(context.err);
             }
         }
         errors.incrementAndGet();
         return null;
     }
 
-    private static boolean addClassToDex(ClassDefItem clazz) {
+    private boolean addClassToDex(ClassDefItem clazz) {
         synchronized (outputDex) {
             outputDex.add(clazz);
         }
@@ -804,7 +806,7 @@ public class Main {
      * @param name {@code non-null;} the fully-qualified internal-form
      * class name
      */
-    private static void checkClassName(String name) {
+    private void checkClassName(String name) {
         boolean bogus = false;
 
         if (name.startsWith("java/")) {
@@ -830,7 +832,7 @@ public class Main {
          * working. Try to help them understand what's happening.
          */
 
-        DxConsole.err.println("\ntrouble processing \"" + name + "\":\n\n" +
+        context.err.println("\ntrouble processing \"" + name + "\":\n\n" +
                 IN_RE_CORE_CLASSES);
         errors.incrementAndGet();
         throw new StopProcessing();
@@ -843,7 +845,7 @@ public class Main {
      * @return {@code null-ok;} the converted {@code byte[]} or {@code null}
      * if there was a problem
      */
-    private static byte[] writeDex(DexFile outputDex) {
+    private byte[] writeDex(DexFile outputDex) {
         byte[] outArray = null;
 
         try {
@@ -865,7 +867,7 @@ public class Main {
                 }
 
                 if (args.statistics) {
-                    DxConsole.out.println(outputDex.getStatistics().toHuman());
+                    context.out.println(outputDex.getStatistics().toHuman());
                 }
             } finally {
                 if (humanOutWriter != null) {
@@ -874,10 +876,10 @@ public class Main {
             }
         } catch (Exception ex) {
             if (args.debug) {
-                DxConsole.err.println("\ntrouble writing output:");
-                ex.printStackTrace(DxConsole.err);
+                context.err.println("\ntrouble writing output:");
+                ex.printStackTrace(context.err);
             } else {
-                DxConsole.err.println("\ntrouble writing output: " +
+                context.err.println("\ntrouble writing output: " +
                                    ex.getMessage());
             }
             return null;
@@ -891,7 +893,7 @@ public class Main {
      * @param fileName {@code non-null;} name of the file
      * @return whether the creation was successful
      */
-    private static boolean createJar(String fileName) {
+    private boolean createJar(String fileName) {
         /*
          * Make or modify the manifest (as appropriate), put the dex
          * array into the resources map, and then process the entire
@@ -912,7 +914,7 @@ public class Main {
                     int length = contents.length;
 
                     if (args.verbose) {
-                        DxConsole.out.println("writing " + name + "; size " + length + "...");
+                        context.out.println("writing " + name + "; size " + length + "...");
                     }
 
                     entry.setSize(length);
@@ -927,10 +929,10 @@ public class Main {
             }
         } catch (Exception ex) {
             if (args.debug) {
-                DxConsole.err.println("\ntrouble writing output:");
-                ex.printStackTrace(DxConsole.err);
+                context.err.println("\ntrouble writing output:");
+                ex.printStackTrace(context.err);
             } else {
-                DxConsole.err.println("\ntrouble writing output: " +
+                context.err.println("\ntrouble writing output: " +
                                    ex.getMessage());
             }
             return false;
@@ -945,7 +947,7 @@ public class Main {
      *
      * @return {@code non-null;} the manifest
      */
-    private static Manifest makeManifest() throws IOException {
+    private Manifest makeManifest() throws IOException {
         byte[] manifestBytes = outputResources.get(MANIFEST_NAME);
         Manifest manifest;
         Attributes attribs;
@@ -981,10 +983,10 @@ public class Main {
      * @param name {@code non-null;} the file name
      * @return {@code non-null;} the opened file
      */
-    private static OutputStream openOutput(String name) throws IOException {
+    private OutputStream openOutput(String name) throws IOException {
         if (name.equals("-") ||
                 name.startsWith("-.")) {
-            return System.out;
+            return context.out;
         }
 
         return new FileOutputStream(name);
@@ -998,14 +1000,14 @@ public class Main {
      *
      * @param stream {@code null-ok;} what to close
      */
-    private static void closeOutput(OutputStream stream) throws IOException {
+    private void closeOutput(OutputStream stream) throws IOException {
         if (stream == null) {
             return;
         }
 
         stream.flush();
 
-        if (stream != System.out) {
+        if (stream != context.out) {
             stream.close();
         }
     }
@@ -1055,13 +1057,13 @@ public class Main {
      * method(s)
      * @param out {@code non-null;} where to dump to
      */
-    private static void dumpMethod(DexFile dex, String fqName,
+    private void dumpMethod(DexFile dex, String fqName,
             OutputStreamWriter out) {
         boolean wildcard = fqName.endsWith("*");
         int lastDot = fqName.lastIndexOf('.');
 
         if ((lastDot <= 0) || (lastDot == (fqName.length() - 1))) {
-            DxConsole.err.println("bogus fully-qualified method name: " +
+            context.err.println("bogus fully-qualified method name: " +
                                fqName);
             return;
         }
@@ -1071,7 +1073,7 @@ public class Main {
         ClassDefItem clazz = dex.getClassOrNull(className);
 
         if (clazz == null) {
-            DxConsole.err.println("no such class: " + className);
+            context.err.println("no such class: " + className);
             return;
         }
 
@@ -1097,7 +1099,7 @@ public class Main {
         }
 
         if (meths.size() == 0) {
-            DxConsole.err.println("no such method: " + fqName);
+            context.err.println("no such method: " + fqName);
             return;
         }
 
@@ -1160,7 +1162,7 @@ public class Main {
     /**
      * A quick and accurate filter for when file path can be trusted.
      */
-    private static class MainDexListFilter implements FileNameFilter {
+    private class MainDexListFilter implements FileNameFilter {
 
         @Override
         public boolean accept(String fullPath) {
@@ -1176,7 +1178,7 @@ public class Main {
     /**
      * A best effort conservative filter for when file path can <b>not</b> be trusted.
      */
-    private static class BestEffortMainDexListFilter implements FileNameFilter {
+    private class BestEffortMainDexListFilter implements FileNameFilter {
 
        Map<String, List<String>> map = new HashMap<String, List<String>>();
 
@@ -1212,7 +1214,7 @@ public class Main {
             }
         }
 
-        private static String getSimpleName(String path) {
+        private String getSimpleName(String path) {
             int index = path.lastIndexOf('/');
             if (index >= 0) {
                 return path.substring(index + 1);
@@ -1246,6 +1248,8 @@ public class Main {
 
         private static final String INPUT_LIST_OPTION = "--input-list";
 
+        public final DxContext context;
+
         /** whether to run in debug mode */
         public boolean debug = false;
 
@@ -1294,6 +1298,9 @@ public class Main {
          */
         public boolean keepClassesInJar = false;
 
+        /** what API level to target */
+        public int minSdkVersion = DexFormat.API_NO_EXTENDED_OPCODES;
+
         /** how much source position info to preserve */
         public int positionInfo = PositionList.LINES;
 
@@ -1342,10 +1349,21 @@ public class Main {
          * mainDexListFile is specified and non empty. */
         public boolean minimalMainDex = false;
 
+        public int maxNumberOfIdxPerDex = DexFormat.MAX_MEMBER_IDX + 1;
+
         /** Optional list containing inputs read in from a file. */
         private List<String> inputList = null;
 
-        private int maxNumberOfIdxPerDex = DexFormat.MAX_MEMBER_IDX + 1;
+        private boolean outputIsDirectory = false;
+        private boolean outputIsDirectDex = false;
+
+        public Arguments(DxContext context) {
+            this.context = context;
+        }
+
+        public Arguments() {
+            this(new DxContext());
+        }
 
         private static class ArgumentsParser {
 
@@ -1448,16 +1466,7 @@ public class Main {
             }
         }
 
-        /**
-         * Parses the given command-line arguments.
-         *
-         * @param args {@code non-null;} the arguments
-         */
-        public void parse(String[] args) {
-            ArgumentsParser parser = new ArgumentsParser(args);
-
-            boolean outputIsDirectory = false;
-            boolean outputIsDirectDex = false;
+        private void parseFlags(ArgumentsParser parser) {
 
             while(parser.getNext()) {
                 if (parser.isArg("--debug")) {
@@ -1480,7 +1489,7 @@ public class Main {
                     statistics = true;
                 } else if (parser.isArg("--optimize-list=")) {
                     if (dontOptimizeListFile != null) {
-                        System.err.println("--optimize-list and "
+                        context.err.println("--optimize-list and "
                                 + "--no-optimize-list are incompatible.");
                         throw new UsageException();
                     }
@@ -1488,7 +1497,7 @@ public class Main {
                     optimizeListFile = parser.getLastValue();
                 } else if (parser.isArg("--no-optimize-list=")) {
                     if (dontOptimizeListFile != null) {
-                        System.err.println("--optimize-list and "
+                        context.err.println("--optimize-list and "
                                 + "--no-optimize-list are incompatible.");
                         throw new UsageException();
                     }
@@ -1508,8 +1517,8 @@ public class Main {
                         jarOutput = false;
                         outputIsDirectDex = true;
                     } else {
-                        System.err.println("unknown output extension: " +
-                                           outName);
+                        context.err.println("unknown output extension: " +
+                                outName);
                         throw new UsageException();
                     }
                 } else if (parser.isArg("--dump-to=")) {
@@ -1528,8 +1537,8 @@ public class Main {
                     } else if (pstr == "lines") {
                         positionInfo = PositionList.LINES;
                     } else {
-                        System.err.println("unknown positions option: " +
-                                           pstr);
+                        context.err.println("unknown positions option: " +
+                                pstr);
                         throw new UsageException();
                     }
                 } else if (parser.isArg("--no-locals")) {
@@ -1550,20 +1559,46 @@ public class Main {
                     maxNumberOfIdxPerDex = Integer.parseInt(parser.getLastValue());
                 } else if(parser.isArg(INPUT_LIST_OPTION + "=")) {
                     File inputListFile = new File(parser.getLastValue());
-                    try{
+                    try {
                         inputList = new ArrayList<String>();
                         readPathsFromFile(inputListFile.getAbsolutePath(), inputList);
-                    } catch(IOException e) {
-                        System.err.println(
+                    } catch (IOException e) {
+                        context.err.println(
                             "Unable to read input list file: " + inputListFile.getName());
                         // problem reading the file so we should halt execution
                         throw new UsageException();
                     }
+                } else if (parser.isArg("--min-sdk-version=")) {
+                    String arg = parser.getLastValue();
+                    int value;
+                    try {
+                        value = Integer.parseInt(arg);
+                    } catch (NumberFormatException ex) {
+                        value = -1;
+                    }
+                    if (value < 1) {
+                        System.err.println("improper min-sdk-version option: " + arg);
+                        throw new UsageException();
+                    }
+                    minSdkVersion = value;
                 } else {
-                    System.err.println("unknown option: " + parser.getCurrent());
+                    context.err.println("unknown option: " + parser.getCurrent());
                     throw new UsageException();
                 }
             }
+        }
+
+
+        /**
+         * Parses all command-line arguments and updates the state of the {@code Arguments} object
+         * accordingly.
+         *
+         * @param args {@code non-null;} the arguments
+         */
+        private void parse(String[] args) {
+            ArgumentsParser parser = new ArgumentsParser(args);
+
+            parseFlags(parser);
 
             fileNames = parser.getRemaining();
             if(inputList != null && !inputList.isEmpty()) {
@@ -1574,11 +1609,11 @@ public class Main {
 
             if (fileNames.length == 0) {
                 if (!emptyOk) {
-                    System.err.println("no input files specified");
+                    context.err.println("no input files specified");
                     throw new UsageException();
                 }
             } else if (emptyOk) {
-                System.out.println("ignoring input files");
+                context.out.println("ignoring input files");
             }
 
             if ((humanOutName == null) && (methodToDump != null)) {
@@ -1586,25 +1621,25 @@ public class Main {
             }
 
             if (mainDexListFile != null && !multiDex) {
-                System.err.println(MAIN_DEX_LIST_OPTION + " is only supported in combination with "
+                context.err.println(MAIN_DEX_LIST_OPTION + " is only supported in combination with "
                     + MULTI_DEX_OPTION);
                 throw new UsageException();
             }
 
             if (minimalMainDex && (mainDexListFile == null || !multiDex)) {
-                System.err.println(MINIMAL_MAIN_DEX_OPTION + " is only supported in combination with "
+                context.err.println(MINIMAL_MAIN_DEX_OPTION + " is only supported in combination with "
                     + MULTI_DEX_OPTION + " and " + MAIN_DEX_LIST_OPTION);
                 throw new UsageException();
             }
 
             if (multiDex && incremental) {
-                System.err.println(INCREMENTAL_OPTION + " is not supported with "
+                context.err.println(INCREMENTAL_OPTION + " is not supported with "
                     + MULTI_DEX_OPTION);
                 throw new UsageException();
             }
 
             if (multiDex && outputIsDirectDex) {
-                System.err.println("Unsupported output \"" + outName +"\". " + MULTI_DEX_OPTION +
+                context.err.println("Unsupported output \"" + outName +"\". " + MULTI_DEX_OPTION +
                         " supports only archive or directory output");
                 throw new UsageException();
             }
@@ -1617,10 +1652,19 @@ public class Main {
         }
 
         /**
-         * Copies relevent arguments over into CfOptions and
-         * DexOptions instances.
+         * Parses only command-line flags and updates the state of the {@code Arguments} object
+         * accordingly.
+         *
+         * @param flags {@code non-null;} the flags
          */
-        private void makeOptionsObjects() {
+        public void parseFlags(String[] flags) {
+            parseFlags(new ArgumentsParser(flags));
+        }
+
+        /**
+         * Copies relevant arguments over into CfOptions and DexOptions instances.
+         */
+        public void makeOptionsObjects() {
             cfOptions = new CfOptions();
             cfOptions.positionInfo = positionInfo;
             cfOptions.localInfo = localInfo;
@@ -1631,12 +1675,13 @@ public class Main {
             cfOptions.statistics = statistics;
 
             if (warnings) {
-                cfOptions.warn = DxConsole.err;
+                cfOptions.warn = context.err;
             } else {
-                cfOptions.warn = DxConsole.noop;
+                cfOptions.warn = context.noop;
             }
 
             dexOptions = new DexOptions();
+            dexOptions.minSdkVersion = minSdkVersion;
             dexOptions.forceJumbo = forceJumbo;
         }
     }
@@ -1645,12 +1690,12 @@ public class Main {
      * Callback class for processing input file bytes, produced by the
      * ClassPathOpener.
      */
-    private static class FileBytesConsumer implements ClassPathOpener.Consumer {
+    private class FileBytesConsumer implements ClassPathOpener.Consumer {
 
         @Override
         public boolean processFileBytes(String name, long lastModified,
                 byte[] bytes)   {
-            return Main.processFileBytes(name, lastModified, bytes);
+            return Main.this.processFileBytes(name, lastModified, bytes);
         }
 
         @Override
@@ -1658,20 +1703,20 @@ public class Main {
             if (ex instanceof StopProcessing) {
                 throw (StopProcessing) ex;
             } else if (ex instanceof SimException) {
-                DxConsole.err.println("\nEXCEPTION FROM SIMULATION:");
-                DxConsole.err.println(ex.getMessage() + "\n");
-                DxConsole.err.println(((SimException) ex).getContext());
+                context.err.println("\nEXCEPTION FROM SIMULATION:");
+                context.err.println(ex.getMessage() + "\n");
+                context.err.println(((SimException) ex).getContext());
             } else if (ex instanceof ParseException) {
-                DxConsole.err.println("\nPARSE ERROR:");
+                context.err.println("\nPARSE ERROR:");
                 ParseException parseException = (ParseException) ex;
                 if (args.debug) {
-                    parseException.printStackTrace(DxConsole.err);
+                    parseException.printStackTrace(context.err);
                 } else {
-                    parseException.printContext(DxConsole.err);
+                    parseException.printContext(context.err);
                 }
             } else {
-                DxConsole.err.println("\nUNEXPECTED TOP-LEVEL EXCEPTION:");
-                ex.printStackTrace(DxConsole.err);
+                context.err.println("\nUNEXPECTED TOP-LEVEL EXCEPTION:");
+                ex.printStackTrace(context.err);
             }
             errors.incrementAndGet();
         }
@@ -1679,13 +1724,13 @@ public class Main {
         @Override
         public void onProcessArchiveStart(File file) {
             if (args.verbose) {
-                DxConsole.out.println("processing archive " + file + "...");
+                context.out.println("processing archive " + file + "...");
             }
         }
     }
 
     /** Callable helper class to parse class bytes. */
-    private static class ClassParserTask implements Callable<DirectClassFile> {
+    private class ClassParserTask implements Callable<DirectClassFile> {
 
         String name;
         byte[] bytes;
@@ -1718,7 +1763,7 @@ public class Main {
      * and then reevaluate the invariant. If there are no further classes in
      * the translation phase, we rotate the dex file.
      */
-    private static class DirectClassFileConsumer implements Callable<Boolean> {
+    private class DirectClassFileConsumer implements Callable<Boolean> {
 
         String name;
         byte[] bytes;
@@ -1817,7 +1862,7 @@ public class Main {
 
 
     /** Callable helper class to translate classes in parallel  */
-    private static class ClassTranslatorTask implements Callable<ClassDefItem> {
+    private class ClassTranslatorTask implements Callable<ClassDefItem> {
 
         String name;
         byte[] bytes;
@@ -1844,7 +1889,7 @@ public class Main {
      * This class is also responsible for coordinating dex file rotation
      * with the DirectClassFileConsumer class.
      */
-    private static class ClassDefItemConsumer implements Callable<Boolean> {
+    private class ClassDefItemConsumer implements Callable<Boolean> {
 
         String name;
         Future<ClassDefItem> futureClazz;
@@ -1892,7 +1937,7 @@ public class Main {
     }
 
     /** Callable helper class to convert dex files in worker threads */
-    private static class DexWriter implements Callable<byte[]> {
+    private class DexWriter implements Callable<byte[]> {
 
         private DexFile dexFile;
 
